import pandas as pd
import pytest
from rdkit import Chem
from rdkit.Chem import AllChem

from m2p import PolyMaker
from m2p.stereo_reactions import (
    get_vinyl_init_dict,
    get_vinyl_prop_dict,
    vinyl_prop_stereo,
    vinyl_terminate_stereo,
)

pm = PolyMaker()


@pytest.fixture
def vinyl():
    return pm.thermoplastic("C=C", DP=10, mechanism="vinyl").smiles_polymer[0]


@pytest.fixture
def RS_ester_reactants():
    poly_name = ["poly(RS-3-hydroxybutyrate)"]
    smiles = ["C[C@@H](O)CC(=O)O.C[C@H](O)CC(=O)O"]

    reactants = pd.DataFrame(
        {"smiles_monomer": smiles, "replicate_structures": [1]},
        columns=["smiles_monomer", "replicate_structures"],
        index=poly_name,
    )
    reactants["monomers"] = reactants.smiles_monomer.apply(
        lambda s: pm.get_monomers(s, stereochemistry=True)
    )
    reactants["mechanism"] = "ester_stereo"

    return reactants


@pytest.fixture
def RS_vinyl_reactants():
    poly_name = ["Polypropylene"]
    smiles = ["CC=C"]

    reactants = pd.DataFrame(
        {"smiles_monomer": smiles, "replicate_structures": [1]}, index=poly_name
    )
    reactants["monomers"] = reactants.smiles_monomer.apply(
        lambda s: pm.get_monomers(s, stereochemistry=True)
    )
    reactants["mechanism"] = "vinyl_stereo"

    return reactants


def calc_pm(smi):
    # Dict to count
    RS_list = []
    mol = Chem.MolFromSmiles(smi)

    for atom in mol.GetAtoms():
        try:
            chiral_center = atom.GetProp("_CIPCode")
            RS_list.append(chiral_center)
        except:
            pass

    # Zip together list to get neighbors and determine R vs M addition
    is_M_add = list(map(lambda t: t[0] == t[1], zip(*(RS_list, RS_list[1:]))))
    pm = sum(is_M_add) / len(is_M_add)

    return pm


def test_vinyl(vinyl):
    assert vinyl == "CCCCCCCCCCCCCCCCCCCC"


def test_ester_stereo_iso(RS_ester_reactants):
    poly_df = pm.thermoplastic_stereo(RS_ester_reactants, DP=10, pm=1, verbose=False)
    pmeso = calc_pm(poly_df["smiles_polymer"][0])

    assert len(poly_df) == 1
    assert pmeso == 1


def test_ester_stereo_syn(RS_ester_reactants):
    poly_df = pm.thermoplastic_stereo(RS_ester_reactants, DP=10, pm=0, verbose=False)
    pmeso = calc_pm(poly_df["smiles_polymer"][0])

    assert len(poly_df) == 1
    assert pmeso == 0


def test_ester_stereo_a(RS_ester_reactants):
    poly_df = pm.thermoplastic_stereo(RS_ester_reactants, DP=10, pm=0.5, verbose=False)
    pmeso = calc_pm(poly_df["smiles_polymer"][0])

    assert len(poly_df) == 1
    assert 0 < pmeso and pmeso < 1


def test_df_pm_ester_stereo(RS_ester_reactants):
    RS_ester_reactants["pm"] = 1

    poly_df = pm.thermoplastic_stereo(RS_ester_reactants, DP=10, verbose=False)
    pmeso = calc_pm(poly_df["smiles_polymer"][0])

    assert len(poly_df) == 1
    assert pmeso == 1


def test_vinyl_enantiomer_dicts():
    def m2s(mol):
        return AllChem.CanonSmiles(AllChem.MolToSmiles(mol))

    smis = ["C=Cc1ccccc1", "C=C([Cl])C"]

    smis_prop_dict = get_vinyl_prop_dict(smis)
    smis_init_dict = get_vinyl_init_dict(smis)

    assert m2s(smis_prop_dict[0]["R"]) == "[Tc]C[C@H]([Re])c1ccccc1"
    assert m2s(smis_prop_dict[0]["S"]) == "[Tc]C[C@@H]([Re])c1ccccc1"

    assert m2s(smis_prop_dict[1]["R"]) == "C[C@](Cl)([Re])C[Tc]"
    assert m2s(smis_prop_dict[1]["S"]) == "C[C@@](Cl)([Re])C[Tc]"

    assert m2s(smis_init_dict[0]["R"]) == "[Re][C@@H](C[At])c1ccccc1"
    assert m2s(smis_init_dict[0]["S"]) == "[Re][C@H](C[At])c1ccccc1"

    assert m2s(smis_init_dict[1]["R"]) == "C[C@](Cl)([Re])C[At]"
    assert m2s(smis_init_dict[1]["S"]) == "C[C@@](Cl)([Re])C[At]"


def test_vinyl_prop():
    smi = ["C=CC"]
    prop_dict = get_vinyl_prop_dict(smi)
    init_dict = get_vinyl_init_dict(smi)

    CIP_assignments = ["R", "S", "R", "S"]

    chain = init_dict[0][CIP_assignments[0]]
    for CIP in CIP_assignments[1:]:
        chain = vinyl_prop_stereo(chain, prop_dict[0][CIP])
    chain = vinyl_terminate_stereo(chain)

    assert AllChem.MolToSmiles(chain) == "CCC[C@H](C)C[C@@H](C)CC(C)C"


def test_vinyl(RS_vinyl_reactants):
    poly_df = pm.thermoplastic_stereo(RS_vinyl_reactants, DP=10, pm=0, verbose=False)
    assert len(poly_df) == 1


def test_vinyl_and_ester(RS_ester_reactants, RS_vinyl_reactants):
    df = pd.concat([RS_vinyl_reactants, RS_ester_reactants]).reset_index(drop=True)
    poly_df = pm.thermoplastic_stereo(df, DP=10, pm=0, verbose=False)
    assert len(poly_df) == 2



"""Functions to generate """
import ast
from random import choices
from typing import List, Union

import pandas as pd
from rdkit.Chem import AllChem

from .utils import get_CIP_assignments, get_monomer_sequences


def get_valid_mols(smi_list):
    mol_list = []
    for smi in smi_list:
        try:
            AllChem.SanitizeMol(smi[0])
            mol_list.append(smi[0])
        except Exception as e:
            print(e)
            pass

    return mol_list


def polyvinyl_stereo(
    reactants: Union[List, str],
    DP: int,
    replicate_structures: int = 1,
    distribution: Union[List[float], List] = [],
    pm: float = 1,
) -> pd.DataFrame:
    """Create polymers ustilizing radical polymerization.

    Parameters
    ----------
    reactants : Union[List, str]
        The smiles to react.
    DP : int
        The degree of polymerization, measured by number of monomers incorporated.
    replicate_structures : int, optional
        How many replicate structures to generate, by default 1
    distribution : Union[List[float], List], optional
        The distribution of monomers determined by their relative abundance. I.e. [1, 2] or [0.33, 0.66] would generate a polymer with two times monomer two than one. Distribution values can be given as any set of numbers, by default []
    pm : float, optional
        The pm value for the polymer, by default 1

    Returns
    -------
    pd.DataFrame
        A dataframe containing the polymerization information of the supplied monomers and conditions.
    """
    try:
        if isinstance(reactants, str):
            reactants = ast.literal_eval(reactants)

        if isinstance(distribution, str):
            distribution = ast.literal_eval(distribution)

        if len(distribution) != len(reactants):
            distribution = [1 for i in reactants]

        if not distribution:
            distribution = [1]

        # first get the list of CIP assignments and monomer sequences to produce
        poly_list = []
        monomer_sequences = get_monomer_sequences(
            replicate_structures, distribution, DP
        )
        CIP_assignments = get_CIP_assignments(replicate_structures, pm, DP)

        prop_dict = get_vinyl_prop_dict(reactants)
        init_dict = get_vinyl_init_dict(reactants)

        for i, CIPs in enumerate(CIP_assignments):
            poly_id = monomer_sequences[i][0]
            poly = init_dict[poly_id][CIPs[0]]
            for j, CIP in enumerate(CIPs[1:]):
                poly_id = monomer_sequences[i][j + 1]
                poly = vinyl_prop_stereo(poly, prop_dict[poly_id][CIP])

            poly = vinyl_terminate_stereo(poly)

            poly_list.append(AllChem.MolToSmiles(poly))

        return_poly = pd.DataFrame(data=poly_list, columns=["smiles_polymer"])

    except BaseException as e:
        return_poly = "ERROR:Vinyl_Stereo_ReactionFailed"

    return return_poly


def get_vinyl_prop_dict(smiles_list: List[str]):
    """Generate the propogation dictionary for radical polymerization."""

    def stm(smi):
        return AllChem.MolFromSmiles(smi)

    enantiomer_dict = dict()
    for i, smi in enumerate(smiles_list):
        # Generate dict entries one by one
        mol = stm(smi)
        rxn_smarts = AllChem.ReactionFromSmarts(
            "[C:1]=[C:2][!#1:3]>>[Tc][C:1][C@@:2]([!#1:3])[Re]"
        )

        products = [
            AllChem.MolFromSmiles(AllChem.MolToSmiles(prod[0]))
            for prod in rxn_smarts.RunReactants((mol,))
        ]

        if len(products) == 1:
            rxn_smarts2 = AllChem.ReactionFromSmarts(
                "[C:1]=[C:2][!#1:3]>>[Tc][C:1][C@:2]([!#1:3])[Re]"
            )

            products.extend(
                [
                    AllChem.MolFromSmiles(AllChem.MolToSmiles(prod[0]))
                    for prod in rxn_smarts2.RunReactants((mol,))
                ]
            )

        enantiomer_dict[i] = {"R": products[0], "S": products[1]}

    return enantiomer_dict


def get_vinyl_init_dict(smi: str):
    """Generate the initiator dictionary for vinyl polymerization."""
    enantiomer_dict = get_vinyl_prop_dict(smi)
    make_terminal = AllChem.ReactionFromSmarts("[Tc][C:1]>>[At][C:1]")

    for key in enantiomer_dict:
        enantiomer_dict[key]["R"] = make_terminal.RunReactants(
            (enantiomer_dict[key]["R"],)
        )[0][0]
        enantiomer_dict[key]["S"] = make_terminal.RunReactants(
            (enantiomer_dict[key]["S"],)
        )[0][0]

    return enantiomer_dict


def vinyl_prop_stereo(poly: AllChem.rdchem.Mol, monomer: AllChem.rdchem.Mol):
    """Carry out a single propogation step."""
    prop_rxn = AllChem.ReactionFromSmarts("[C:1][Re].[Tc][C:2]>>[C:1][C:2]")

    products = prop_rxn.RunReactants(
        (
            poly,
            monomer,
        )
    )

    return products[0][0]


def vinyl_terminate_stereo(poly: AllChem.rdchem.Mol):
    """Carry out a single termination step."""
    term_rxn = AllChem.ReactionFromSmarts("([Re][C:1].[At][C:2])>>([C:1].[C:2])")

    products = term_rxn.RunReactants((poly,))

    return products[0][0]


# #############
# Polyesters
def polyester_stereo(
    reactants: Union[List, str],
    DP: int,
    replicate_structures: int = 1,
    distribution: Union[List[float], List] = [1],
    pm: float = 1,
):
    """Create polyesters with stereochemical information.

    Parameters
    ----------
    reactants : Union[List, str]
        The smiles to react.
    DP : int
        The degree of polymerization, measured by number of monomers incorporated.
    replicate_structures : int, optional
        How many replicate structures to generate, by default 1
    distribution : Union[List[float], List], optional
        The distribution of monomers determined by their relative abundance. I.e. [1, 2] or [0.33, 0.66] would generate a polymer with two times monomer two than one. Distribution values can be given as any set of numbers, by default []
    pm : float, optional
        The pm value for the polymer, by default 1

    Returns
    -------
    pd.DataFrame
        A dataframe containing the polymerization information of the supplied monomers and conditions.
    """
    try:
        if isinstance(reactants, str):
            reactants = ast.literal_eval(reactants)

        if isinstance(distribution, str):
            distribution = ast.literal_eval(distribution)

        if not distribution:
            distribution = []

        # Get the smiles into correct form
        reactants = [[x, y] for x, y in zip(reactants[::2], reactants[1::2])]

        # first get the list of CIP assignments and monomer sequences to produce
        poly_list = []
        monomer_sequences = get_monomer_sequences(
            replicate_structures, distribution, DP
        )
        CIP_assignments = get_CIP_assignments(replicate_structures, pm, DP)

        # Next generate prop and init dicts
        prop_dict = get_ester_prop_dict(reactants)
        init_dict = get_ester_init_dict(reactants)

        for i, CIPs in enumerate(CIP_assignments):
            # Init
            poly_id = monomer_sequences[i][0]
            poly = init_dict[poly_id][CIPs[0]]

            for j, CIP in enumerate(CIPs[1:]):
                poly_id = monomer_sequences[i][j + 1]
                poly = ester_prop_stereo(poly, prop_dict[poly_id][CIP])

            poly = ester_terminate_stereo(poly)

            poly_list.append(AllChem.MolToSmiles(poly))

        return_poly = pd.DataFrame(data=poly_list, columns=["smiles_polymer"])

    except BaseException as e:
        return_poly = "ERROR:Ester_Stereo_ReactionFailed"

    return return_poly


def get_ester_prop_dict(smiles_list: str):
    """Generate the propogation dictionary. Contains monomers + enantiomers."""
    # Xe reacts with Argon
    def stm(smi):
        return AllChem.MolFromSmiles(smi)

    rxn_smarts = AllChem.ReactionFromSmarts(
        "([O;D1:1][C:2]=[O:3].[C:4][O;D1:5])>>([Re:1][C:2]=[O:3].[C:4][Bh:3])"
    )

    enantiomer_dict = dict()
    for i, monomers in enumerate(smiles_list):
        products = [rxn_smarts.RunReactants((stm(smi),))[0][0] for smi in monomers]

        enantiomer_dict[i] = {"R": products[0], "S": products[1]}

    return enantiomer_dict


def get_ester_init_dict(smiles_list: str):
    """Generate the initiation dictionary. Contains monomers + enantiomers."""
    # Pb is terminal on the carboxylic acid, doesn't react
    # Xe reacts with Argon
    def stm(smi):
        return AllChem.MolFromSmiles(smi)

    rxn_smarts = AllChem.ReactionFromSmarts(
        "([O;D1:1][C:2]=[O:3].[C:4][O;D1:5])>>([Tc:1][C:2]=[O:3].[C:4][Bh:3])"
    )

    enantiomer_dict = dict()
    for i, monomers in enumerate(smiles_list):
        products = [rxn_smarts.RunReactants((stm(smi),))[0][0] for smi in monomers]

        enantiomer_dict[i] = {"R": products[0], "S": products[1]}

    return enantiomer_dict


def ester_prop_stereo(poly: AllChem.rdchem.Mol, monomer: AllChem.rdchem.Mol):
    prop_rxn = AllChem.ReactionFromSmarts("[C:1][Bh].[Re][C:2]>>[C:1][O][C:2]")

    products = prop_rxn.RunReactants(
        (
            poly,
            monomer,
        )
    )

    return products[0][0]


def ester_terminate_stereo(poly: AllChem.rdchem.Mol):
    prop_rxn = AllChem.ReactionFromSmarts("([Tc][C:1].[Bh][C:2])>>([O][C:1].[O][C:2])")

    products = prop_rxn.RunReactants((poly,))

    return products[0][0]



# import pytest
# from rdkit import Chem

# from m2p import Monomer


# def test_acid_ol_monomer():
#     acid_ol_smiles = "O=C(O)CC(C)CCO"
#     m = Monomer(acid_ol_smiles)
#     assert m.molecular_weight == 132.078644244
#     assert m.canonical_smiles == Chem.CanonSmiles(acid_ol_smiles)
#     assert m.esterification_enantiomers == [
#         ("C[C@@H](CCO)CC(=O)O", "C[C@H](CCO)CC(=O)O")
#     ]


# def test_acid_ol_multiple_stereo():
#     acid_ol_smiles = "O=C(O)CC(C)C(C)C(C)CO"
#     m = Monomer(acid_ol_smiles)

#     assert m.esterification_enantiomers == [
#         ("CC(CC(=O)O)C(C)[C@@H](C)CO", "CC(CC(=O)O)C(C)[C@H](C)CO"),
#         ("CC(CO)C(C)[C@H](C)CC(=O)O", "CC(CO)C(C)[C@@H](C)CC(=O)O"),
#         ("CC(CO)[C@H](C)C(C)CC(=O)O", "CC(CO)[C@@H](C)C(C)CC(=O)O"),
#     ]



"""Monomer class with parameters and methods."""

from typing import Tuple, Union

import pandas as pd
from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt
from rdkit.Chem.rdchem import Mol

functionality_smarts = {
    "ols": "[C,c;!$(C=O)][OH]",
    "aliphatic_ols": "[C;!$(C=O);!$([a])][OH]",
    "acids": "[#6][#6](=[#8:4])([F,Cl,Br,I,#8H,O-])",
    "prime_amines": "[#6;!$(C=O)][NH2;!$([NH2+])]",
    "carbonates": "[O]=[C]([F,Cl,Br,I,O])([F,Cl,Br,I,O])",
    "acidanhydrides": "[#8]([#6](=[#8]))([#6](=[#8]))",
    "prime_thiols": "[#6;!$(C=O)][SH]",
}


def molecule_from_smiles(smiles: str) -> Union[Mol, None]:
    """Generate rdkit mol from smiles

    Parameters
    ----------
    smiles : str
        SMILES string

    Returns
    -------
    Union[Mol, None]
        RDKit Molecule, or None if can't generate
    """
    try:
        mol = Chem.MolFromSmiles(smiles)
    except:
        mol = None

    return mol


def get_functionality(reactants, distribution=[]):
    """gets the functional groups from a list of reactants

    inputs: list of smiles
    output: dataframe with count of functional groups
    """

    def id_functionality(r):
        mol = Chem.MolFromSmiles(r.name)
        r.ols = len(
            mol.GetSubstructMatches(Chem.MolFromSmarts(functionality_smarts["ols"]))
        )
        r.aliphatic_ols = len(
            mol.GetSubstructMatches(
                Chem.MolFromSmarts(functionality_smarts["aliphatic_ols"])
            )
        )
        r.acids = len(
            mol.GetSubstructMatches(Chem.MolFromSmarts(functionality_smarts["acids"]))
        )
        r.prime_amines = len(
            mol.GetSubstructMatches(
                Chem.MolFromSmarts(functionality_smarts["prime_amines"])
            )
        )
        r.carbonates = len(
            mol.GetSubstructMatches(
                Chem.MolFromSmarts(functionality_smarts["carbonates"])
            )
        )
        r.acidanhydrides = len(
            mol.GetSubstructMatches(
                Chem.MolFromSmarts(functionality_smarts["acidanhydrides"])
            )
        )
        return r

    df_func = pd.DataFrame(
        data=0,
        index=reactants,
        columns=[
            "ols",
            "acids",
            "prime_amines",
            "carbonates",
            "aliphatic_ols",
            "acidanhydrides",
        ],
    )
    df_func = df_func.apply(lambda r: id_functionality(r), axis=1)

    # appends distribution to dataframe

    if len(distribution) == 0:
        df_func["distribution"] = [1] * df_func.shape[0]
    else:
        df_func["distribution"] = list(distribution)
    return df_func


def enumerate_ester_enantiomers(smiles: str) -> Tuple[Tuple[str]]:
    """Generate enantiomer pairs for a monomer that would participate in an esterification reaction

    Parameters
    ----------
    smiles : str
        SMILES string of the monomer

    Returns
    -------
    Tuple[Tuple[str]]
        The enantiomer pairs, or a single string if no enantiomers are created
    """
    try:
        mol = Chem.MolFromSmiles(smiles)
        # Get the atom ids for the acid and ol functionalities
        acid = Chem.MolFromSmarts("[O]-[C]=[O]")
        ol = Chem.MolFromSmarts("[C;X4]-[O]")
        acid_atoms = mol.GetSubstructMatch(acid)
        ol_atoms = mol.GetSubstructMatch(ol)

        # Get the carbon atoms for both, doesn't matter for acid, but does for ol
        atoms = mol.GetAtoms()

        for atom_i in acid_atoms:
            if atoms[atom_i].GetAtomicNum() == 8:
                acid_O = atom_i

        for atom_i in ol_atoms:
            if atoms[atom_i].GetAtomicNum() == 8:
                ol_O = atom_i

        # Get shortest path (backbone) and make stereo based on that
        # Does shortest path make sense?
        atom_path = Chem.GetShortestPath(mol, acid_O, ol_O)

        # Get sites that can be stereo
        stereo_sites = [
            sinfo.centeredOn for sinfo in Chem.rdmolops.FindPotentialStereo(mol)
        ]

        # Find the enantiomers based on stereo sites within the shortest path
        enantiomers = []

        for atom_i in set(atom_path).intersection(stereo_sites):
            atoms[atom_i].SetChiralTag(Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW)
            smilesR = Chem.MolToSmiles(mol)
            atoms[atom_i].SetChiralTag(Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW)
            smilesS = Chem.MolToSmiles(mol)

            enantiomers.append((smilesR, smilesS))
            atoms[atom_i].SetChiralTag(Chem.rdchem.ChiralType.CHI_UNSPECIFIED)

    except BaseException as e:
        enantiomers = [[smiles]]

    if not enantiomers:
        enantiomers = [[smiles]]

    return enantiomers


class Monomer:
    def __init__(self, smiles: str):
        self.smiles = smiles
        self.canonical_smiles = Chem.CanonSmiles(smiles)
        self.molecule = molecule_from_smiles(self.smiles)
        self.molecular_weight = ExactMolWt(self.molecule)

        df = pd.DataFrame([self.smiles], columns=["smiles"])
        self.functionality = get_functionality(df.smiles)

    def __repr__(self) -> str:
        if self.is_valid:
            return f"Valid Monomer with smiles {self.smiles}"
        else:
            return f"Invalid Monomer with smiles {self.smiles}"

    @property
    def esterification_enantiomers(self) -> Tuple[Tuple[str]]:
        """Get possible enantiomers that would participate in an esterification reaction

        Returns
        -------
        Tuple[Tuple[str, str]]
            Tuple containing diads of enantiomers
        """
        return enumerate_ester_enantiomers(self.smiles)
